@@ -1,4 +1,8 @@
 
+import urllib
+import sys
+
+import configparser
 from ipaddress import ip_network, ip_address
 import select
 import threading
@@ -22,52 +26,47 @@ from nacl.hash import blake2b
 import nacl
 import hashlib
 import stat
-
+import struct
 from . import util
-
+from .cachingproxy import CachingProxy
 
 services = weakref.WeakValueDictionary()
 
 socket.setdefaulttimeout(5)
 
-# Todo fix this
-
-# def createWifiChecker():
-
-#     """Detects if we are on something like WiFi.  But if we aren't on android at all, assume that we are always connected to a LAN, even htough there are 4g laptops.
-#        don't use for anything critical because of that.
-
-#     """
-#     def alwaysTrue():
-#         return True
-#     try:
-#         from kivy.utils import platform
-#         from jnius import autoclass
-#     except:
-#         return alwaysTrue
-
-#     if platform != 'android':
-#             return alwaysTrue
 
 
-#     def check_connectivity():
+def createWifiChecker():
+    """Detects if we are on something like WiFi, and if we have sufficient battery power to keep going.
+
+    But if we aren't on android at all, assume that we are always connected to a LAN, even htough there are 4g laptops.
+    don't use for anything critical because of that.
+
+    """
+    def alwaysTrue():
+        return True
+
+    try:
+        import kivy.utils
+    except ImportError:
+        return alwaysTrue
+
+    if kivy.utils.platform != 'android':
+        return alwaysTrue
+
+    from plyer import wifi
+    from plyer import battery
+
+    def check_connectivity():
+        if battery.status['isCharging'] or battery.status['percentage'] > 30:
+            return wifi.is_connected()
+        else:
+            return False
+
+    return check_connectivity()
 
 
-#         Activity = autoclass('android.app.Activity')
-#         PythonActivity = autoclass('org.kivy.android.PythonActivity')
-#         activity = PythonActivity.mActivity
-#         ConnectivityManager = autoclass('android.net.ConnectivityManager')
-
-#         con_mgr = activity.getSystemService(Activity.CONNECTIVITY_SERVICE)
-
-#         conn = con_mgr.getNetworkInfo(ConnectivityManager.TYPE_WIFI).isConnectedOrConnecting()
-#         if conn:
-#             return True
-
-#     return check_connectivity()
-
-# isOnLan= createWifiChecker()
-def isOnLan(): return True
+isOnLan = createWifiChecker()
 
 
 P2P_PORT = 7009
@@ -90,7 +89,25 @@ try:
     from android.storage import app_storage_path
     settings_path = app_storage_path()
 except:
-    settings_path = '~/.hardlinep2p/'
+    settings_path = os.path.expanduser('~/.hardlinep2p/')
+
+
+try:
+    from jnius import autoclass, cast
+    PythonActivity = autoclass('org.kivy.android.PythonActivity')
+    Environment = autoclass('android.os.Environment')
+    context = cast('android.content.Context', PythonActivity.mActivity)
+
+    r = context.getExternalFilesDir(
+        Environment.getDataDirectory().getAbsolutePath()
+    ).getAbsolutePath()
+
+    user_services_dir = os.path.join(r, "services")
+    proxy_cache_root = os.path.join(r, "proxycache")
+
+except:
+    user_services_dir = os.path.expanduser('~/.hardlinep2p/services/')
+    proxy_cache_root = os.path.expanduser('~/.hardlinep2p/proxycache/')
 
 # clientidkeyfile = os.path.join(os.path.expanduser(settings_path), "clientid.txt")
 
@@ -234,8 +251,15 @@ class DiscoveryCache():
 
             self.lastTriedDHT = time.time()
 
+            # Rolling code changes the DHT key every 24 hours, ensuring that we don't heavily load down any particular
+            # DHT node for more than a day, if there is somehow an incredibly popular site.
+            # It also gives less information to people who don't know the unhashed ID, who may want to
+            # spy on when your service is up, or some crap like that.
+            timePeriod = struct.pack("<Q", int(time.time()/(3600*24)))
+            rollingCode = blake2b(bytes.fromhex(self.infohash)+timePeriod,
+                                  encoder=nacl.encoding.RawEncoder())[:20]
             # Use SHA1 here as it is openDHT custom
-            k = hashlib.sha1(self.infohash.encode()).digest()[:20].hex()
+            k = hashlib.sha1(rollingCode.hex().encode()).digest()[:20].hex()
 
             r = None
             lines = []
@@ -272,21 +296,20 @@ class DiscoveryCache():
 
     def get(self, invalidate=False):
 
-        # Not on WiFi, refrain from trying and failing to do multicast discovery.
-        if isOnLan():
-            if invalidate:
-                self.LPDcacheRecord = (None, 0, 10, '')
-            # Try a local serach
-            for i in range(0, 4):
+        if invalidate:
+            self.LPDcacheRecord = (None, 0, 10, '')
 
-                # Look in the cache first
-                x = self.LPDcacheRecord
-                if x[1] > (time.time() - 60):
-                    return [x[0]]
+        # Try a local serach
+        for i in range(0, 4):
 
-                self.doLookup()
-                # Wait a bit for replies, waiting a little longer every time
-                time.sleep(i*0.1 + 0.05)
+            # Look in the cache first
+            x = self.LPDcacheRecord
+            if x[1] > (time.time() - 60):
+                return [x[0]]
+
+            self.doLookup()
+            # Wait a bit for replies, waiting a little longer every time
+            time.sleep(i*0.1 + 0.05)
 
         # Local search failed, we haven't seen a packet from them, so we are probably not on their network.
         # lets try a stored WAN address search, maybe we have a record we can use?
@@ -389,12 +412,16 @@ connections = weakref.WeakValueDictionary()
 
 
 class Service():
-    def __init__(self, cert, destination, port, info={'title': ''},):
+    def __init__(self, cert, destination, port, info={'title': ''}, friendlyName=None, cacheSettings={}):
         global P2P_PORT
 
         self.certfile = cert
         self.dest = destination+":"+str(port)
 
+        self.closed = False
+
+        # This is a name we use in GUI service listings for display
+        self.friendlyName = friendlyName
         # #Used for tracking who we have directly connected to on the lan.
         # #For GDPR reasons, we do not store data besides the anonymous ID and the fact that they connected at some point.
         # self.lanClientsFile = self.certfile+".lanclients"
@@ -426,24 +453,94 @@ class Service():
         else:
             self.cert_gen(cert)
 
+        # Should there be a proxy dir set up, we run an in-process caching proxy
+        # Just for that.  It's mostly an extra convenience and not something
+        # Deeply integrated.
+        self.cachingProxy = None
+        proxyDir = cacheSettings.get("directory")
+        if proxyDir:
+            proxyDir = os.path.join(proxy_cache_root, proxyDir)
+
+            # No spurious port 80 that isn't needed and breaks stuff
+            if port and not int(port) == 80:
+                dest = destination+":"+str(port)
+            else:
+                dest = destination
+            yes = ('yes','true','on','Yes','True','On','TRUE')
+            self.cachingProxy = CachingProxy(dest, proxyDir, maxAge=cacheSettings.get("maxAge"), 
+            maxSize=cacheSettings.get("maxSize"),
+            downloadRateLimit=cacheSettings.get("downloadRateLimit"), 
+            allowListing=cacheSettings.get("allowListing") in yes,
+            dynamicContent=cacheSettings.get("dynamicContent") in yes)
+
+            for i in range(0, 10):
+                if self.cachingProxy.port:
+                    break
+                time.sleep(1)
+            else:
+                raise RuntimeError("Proxy did not start")
+
+            self.dest = "localhost:"+str(self.cachingProxy.port)
+
         services[self.keyhash.hex()] = self
 
         self.lpd = util.LPDPeer("HARDLINE-SERVICE", "HARDLINE-SEARCH")
+        self.lpd.register(self.password+"-"+self.keyhash.hex(),
+                          WANPortContainer, info)
 
-        self.lpd.register(self.password+"-"+self.keyhash.hex(), P2P_PORT, info)
+    def close(self):
+        global P2P_PORT
+        try:
+            services.pop(self.keyhash.hex())
+        except KeyError:
+            pass
+
+        if self.cachingProxy:
+            try:
+                self.cachingProxy.server.shutdown()
+                self.cachingProxy = None
+            except KeyError:
+                pass
+
+        self.closed = True
+
+        self.lpd.unregister(self.password+"-" +
+                            self.keyhash.hex())
 
     def dhtPublish(self, node):
         # Publish this service to the DHT for WAN discovery.
 
+        timePeriod = struct.pack("<Q", int(time.time()/(3600*24)))
+        rollingCode = blake2b(self.keyhash+timePeriod,
+                              encoder=nacl.encoding.RawEncoder())[:20]
+
         # We never actually know if this will be available on the platform or not
-        try:
-            import opendht as dht
-        except:
+        if node:
+            try:
+                import opendht as dht
+
+                with dhtlock:
+                    node.put(dht.InfoHash.get(rollingCode.hex()),
+                             dht.Value(getWanHostsString().encode()))
+            except Exception:
+                print("Could not use local DHT node")
+                print(traceback.format_exc())
             return
 
-        with dhtlock:
-            node.put(dht.InfoHash.get(self.keyhash.hex()),
-                     dht.Value(getWanHostsString().encode()))
+        # Using a DHT proxy we can host a site without actually using the DHT directly.
+        # This is for future direct-from-android hosting.
+        for i in DHT_PROXIES:
+            import requests
+            try:
+                data = {"data": base64.b64encode(
+                    getWanHostsString()).decode(), "id": "id 1", "seq": 0, "type": 3}
+                url = i+hashlib.sha1(rollingCode.hex().encode()
+                                     ).digest()[:20].hex()
+                r = requests.post(url, data=data)
+                r.raise_for_status()
+                break
+            except Exception:
+                print(traceback.format_exc())
 
     def handleConnection(self, sock):
         "Handle incoming encrypted connection from another hardline instance.  The root server code has alreadt recieved the SNI and has dispatched it to us"
@@ -479,7 +576,7 @@ class Service():
                 raise RuntimeError(
                     "Too many incoming P2P connections all at once")
 
-            h, p = self.dest.split(":")
+            h, p = self.dest.split(":")[-2:]
             conn.connect((h, int(p)))
 
             # Wait for ready
@@ -505,6 +602,8 @@ class Service():
 
             # The first part of the data is reserved for an OOB header
             while(1):
+                if self.closed:
+                    raise RuntimeError("Closed at server side")
                 r, w, x = select.select([sock, ], [], [], 1)
                 if r:
                     oob += sock.recv(4096)
@@ -523,6 +622,8 @@ class Service():
             conn.send(overflow)
 
             while(1):
+                if self.closed:
+                    raise RuntimeError("Closed at server side")
                 r, w, x = select.select([sock, conn], [], [], 1)
 
                 # Whichever one has data, shove it down the other one
@@ -682,34 +783,38 @@ def server_thread(sock):
         # Use fullservice for discovery, the LPD code takes care of not actually sending the password part
         hosts = discover(fullservice.decode())
 
-        # We do our own verification
-        sk = ssl.create_default_context()
+        conn = None
 
-        sk.options |= ssl.OP_NO_TLSv1
-        sk.options |= ssl.OP_NO_TLSv1_1
+        def connect(to):
+            # We do our own verification
+            sk = ssl.create_default_context()
 
-        sk.check_hostname = False
-        sk.verify_mode = ssl.CERT_NONE
-        sock2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+            sk.options |= ssl.OP_NO_TLSv1
+            sk.options |= ssl.OP_NO_TLSv1_1
 
-        # clientID = blake2b(clientUUID.encode()+service ,encoder=nacl.encoding.RawEncoder())[:20]
+            sk.check_hostname = False
+            sk.verify_mode = ssl.CERT_NONE
+            sock2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
 
-        # Use TCP keepalives here.  Note that this isn't really secure because it's at the TCP layer,
-        # Someone could DoS it, but DoS is hard to stop no matter what.
-        sock2.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
-        sock2.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 5)
-        sock2.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 15)
-        sock2.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)
+            # clientID = blake2b(clientUUID.encode()+service ,encoder=nacl.encoding.RawEncoder())[:20]
 
-        conn = sk.wrap_socket(sock2, server_hostname=service)
+            # Use TCP keepalives here.  Note that this isn't really secure because it's at the TCP layer,
+            # Someone could DoS it, but DoS is hard to stop no matter what.
+            sock2.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
+            sock2.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 5)
+            sock2.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 15)
+            sock2.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)
 
+            conn = sk.wrap_socket(sock2, server_hostname=service)
+            conn.connect(to)
+            return conn
         try:
             # Try our discovered hosts
             for host in hosts:
                 try:
                     connectingTo = host
                     print("trying", host)
-                    conn.connect(host)
+                    conn = connect(host)
                     break
                 except:
                     print(traceback.format_exc())
@@ -719,7 +824,7 @@ def server_thread(sock):
                     try:
                         connectingTo = host
                         print("trying", host)
-                        conn.connect(host)
+                        conn = connect(host)
                         break
                     except:
                         print(traceback.format_exc())
@@ -729,7 +834,7 @@ def server_thread(sock):
                         try:
                             print("trying", host)
                             connectingTo = host
-                            conn.connect(host)
+                            conn = connect(host)
                             break
                         except:
                             print(traceback.format_exc())
@@ -737,8 +842,8 @@ def server_thread(sock):
                         raise RuntimeError(
                             "All saved host options and dht options failed:"+str(hosts))
         except:
-            conn.close()
-            sock2.close()
+            if conn:
+                conn.close()
             raise
        # else:
             # We have failed, now we have to use DHT lookup
@@ -890,26 +995,28 @@ def taskloop():
 
         # If we can't get a UPNP listing we have to rely on a manual port mapping.
         if not success:
-            if time.time() > lastUsedIPAPI-30*40:
-                lastUsedIPAPI = time.time()
-                try:
-                    import requests
-                    r = requests.get("https://api.ipify.org/", timeout=5)
-                    r.raise_for_status()
-                    ExternalAddrs[0] = r.text
-                    success = True
-                except:
-                    print(traceback.format_exc())
+            # Don't hammer IP API for no reason if we don't actually have services to serve.
+            if services and isOnLan():
+                if time.time() > lastUsedIPAPI-40*60:
+                    lastUsedIPAPI = time.time()
+                    try:
+                        import requests
+                        r = requests.get("https://api.ipify.org/", timeout=5)
+                        r.raise_for_status()
+                        ExternalAddrs[0] = r.text
+                        success = True
+                    except:
+                        print(traceback.format_exc())
 
-        if not success:
-            ExternalAddrs[0] = ''
+                if not success:
+                    ExternalAddrs[0] = ''
 
-        if dhtContainer[0]:
-            try:
-                for i in services:
+        try:
+            for i in services:
+                if isOnLan():
                     services[i].dhtPublish(dhtContainer[0])
-            except:
-                print(traceback.format_exc())
+        except:
+            print(traceback.format_exc())
 
         time.sleep(8*60)
 
@@ -917,38 +1024,69 @@ def taskloop():
 portMapping = None
 
 running = True
+exited = True
 
 
 def stop():
     # Will not instant;y stop active connections
-    global running
-    running = False
+    global running, exited
+    while not exited:
+        running = False
+        time.sleep(0.1)
 
 
-def start(localport):
-    global P2P_PORT, WANPort, portMapping, running
+cached_localport = 7008
+
+
+def start(localport=None):
+    global P2P_PORT, WANPort, portMapping, running, exited, cached_localport
 
     running = True
-    # This is the server context we use for localhost coms
 
-    bindsocket = socket.socket()
-    if localport:
-        #Try for 30 seconds of waiting to see if the port becomes available.
-        for i in range(30):
-            try:
-                bindsocket.bind(('localhost', localport))
-                bindsocket.listen()
-                break
-            except OSError:
-                if i<9:
-                    time.sleep(1)
-                else:
-                    raise
+    ports = []
+
+    localport = localport or cached_localport
+    cached_localport = localport
+
+    try:
+        # This is the server context we use for localhost coms
+        bindsocket = socket.socket()
+        if localport:
+            # Try for 30 seconds of waiting to see if the port becomes available.
+            for i in range(30):
+                try:
+                    bindsocket.bind(('localhost', localport))
+                    bindsocket.listen()
+                    ports.append(bindsocket)
+                    break
+                except OSError:
+                    if i < 9:
+                        time.sleep(1)
+                    else:
+                        raise
+    except Exception:
+        bindsocket = None
+        print("Failed to start localhost gateway.  Perhaps another Hardline instance is running.  Continuing as server only")
 
     # This is the server context we use for the remote coms, accepting incoming ssl connections from other instances and proxying them into
     # local services
 
     p2p_bindsocket = socket.socket()
+
+    for i in range(30):
+        try:
+            p2p_bindsocket.bind(('0.0.0.0', P2P_PORT))
+            p2p_bindsocket.listen()
+            WANPortContainer[0] = P2P_PORT
+            break
+        except OSError:
+            if i < 28:
+                time.sleep(1)
+                P2P_PORT += 1
+
+            else:
+                raise
+
     if services:
         from . import upnpwrapper
 
@@ -969,19 +1107,6 @@ def start(localport):
             node.bootstrap("bootstrap.jami.net", "4222")
             dhtContainer[0] = node
 
-
-        for i in range(30):
-            try:
-                p2p_bindsocket.bind(('0.0.0.0', P2P_PORT))
-                p2p_bindsocket.listen()
-                break
-            except OSError:
-                if i<9:
-                    time.sleep(1)
-                else:
-                    raise
-
- 
         # Only daemons exposing a service need a WAN mapping
         t = threading.Thread(target=taskloop, daemon=True)
         t.start()
@@ -991,7 +1116,7 @@ def start(localport):
         # Default to the p2p port
         WANPortContainer[0] = P2P_PORT
 
-        for i in range(0, 25):
+        for i in range(0, 5):
             try:
                 portMapping = upnpwrapper.addMapping(
                     P2P_PORT, "TCP", WANPort=WANPortContainer[0])
@@ -1005,9 +1130,14 @@ def start(localport):
             print("Failed to register port mapping, you will need to manually configure.")
             WANPortContainer[0] = P2P_PORT
 
+    toScan = [p2p_bindsocket]
+
+    if bindsocket:
+        toScan.append(bindsocket)
+
+    exited = False
     while(running):
-        r, w, x = select.select([bindsocket, p2p_bindsocket] if services else [
-                                bindsocket], [], [], 1)
+        r, w, x = select.select(toScan, [], [], 1)
         try:
 
             # Whichever one has data, shove it down the other one
@@ -1017,8 +1147,136 @@ def start(localport):
                 else:
                     if services:
                         handleP2PClient(i.accept()[0])
+                    else:
+                        i.accept()[0].close()
         except:
             pass
 
-    bindsocket.close()
-    p2p_bindsocket.close()
+    try:
+        bindsocket.close()
+        p2p_bindsocket.close()
+    finally:
+        exited = True
+
+
+def closeServices(only=None):
+    "Only lets us only close one,  by it's friendly name"
+    for i in list(services.values()):
+        try:
+            if only:
+                if not i.friendlyName == only:
+                    continue
+            i.close()
+        except:
+            print(traceback.format_exc())
+
+
+userServices = {}
+
+# "User Services" are configurable services stored in files, asw opposed to those defined in code
+
+
+def loadUserServices(serviceDir, only=None):
+    "Load services from a configuration directory.  Only to  only reload one."
+    try:
+        os.makedirs(serviceDir)
+    except:
+        pass
+
+    if os.path.exists(serviceDir):
+        x = os.listdir(serviceDir)
+        for i in x:
+            if not i.endswith(".ini"):
+                continue
+            if only:
+                if not i == only+".ini":
+                    continue
+            try:
+                config = configparser.ConfigParser()
+                config.read(os.path.join(serviceDir, i))
+
+                if "Info" in config.sections():
+                    title = config['Info'].get("title", 'untitled')
+                else:
+                    title = 'untitled'
+
+                if "Cache" in config.sections():
+                    cache = config['Cache']
+                else:
+                    cache = {}
+
+                service = config['Service']
+
+                # Close any existing service by that same friendly local name
+                closeServices(i[:-4])
+
+                certFile = os.path.join(serviceDir, i+".cert")
+                if service.get('certfile',''):
+                    certFile=service['certfile']
+
+                # Take friendly name from filename
+                s = Service(certFile, service['service'], int(
+                    service.get('port', '80')), {'title': title}, friendlyName=i[:-4], cacheSettings=cache)
+                print("Serving a service from "+service['service'])
+
+                userServices[i] = s
+            except:
+                print(traceback.format_exc())
+
+
+def makeUserService(dir, name, title="A service", service="localhost", port='80',  certfile=None, cacheInfo={}):
+    c = configparser.ConfigParser()
+    c.add_section("Service")
+    c.add_section("Info")
+    c.add_section("Cache")
+
+    file = os.path.join(dir, name+'.ini')
+
+    sinfo = c['Service']
+    sinfo['port'] = str(port)
+    sinfo['service'] = service
+    # Default gets used
+    sinfo['certfile'] = ""
+
+    info = c['Info']
+    info['title'] = title
+
+    for i in cacheInfo:
+        c['Cache'][i] = cacheInfo[i]
+
+    with open(file, "w") as f:
+        c.write(f)
+    loadUserServices(dir, name)
+
+
+def delUserService(dir, name):
+
+    file = os.path.join(dir, name+'.ini')
+
+    if os.path.exists(dir):
+        for n in os.listdir(dir):
+            i = os.path.join(dir, n)
+            if file and i.startswith(file):
+                os.remove(i)
+    closeServices(dir, name)
+
+
+def listServices(serviceDir):
+    try:
+        os.makedirs(serviceDir)
+    except:
+        pass
+
+    services = {}
+    if os.path.exists(serviceDir):
+        for i in os.listdir(serviceDir):
+            if not i.endswith(".ini"):
+                continue
+            try:
+                config = configparser.ConfigParser()
+                config.read(os.path.join(serviceDir, i))
+                services[i[:-4]] = config
+            except:
+                print(traceback.format_exc())
+
+    return services
